"""MALDI mass list reader module.

This module provides functionality to read and parse mass spectrometry data from MALDI Excel exports through the pibin
FileReader class. It handles multi-sheet xlsx files containing mass lists with specific column headers for m/z values,
intensities, and other mass spectrometry parameters found in COLUMN_HEADERS below.
"""

from typing import Dict
import os
import pandas as pd

from piblin.dataio import FileReader
from piblin.data import Measurement, MeasurementSet

from ..fmr_classes.fmr_datasets import FractionalMRDataset
from .. import fmr_parameters as p

COLUMN_HEADERS = ['m/z', 'time', 'Intens.', 'SN', 'Quality Fac.', 'Res.', 'Area', 'Rel. Intens.', 'FWHM', 'Chi^2',
                  'Bk. Peak']


class MaldiMassListReader(FileReader):
    """
    Reads .xlsx files from maldi export. Reader for .xlsx document where the sheets are named based on the sample
    information.
    """
    supported_extensions = {'xlsx'}

    def _read_file_contents(self, filepath: str, **read_kwargs):
        """Read the contents of a file into a single object.

        Parameters
        ----------
        filepath : str
            The path to the file to be read.
        read_kwargs : dict
            Read parameters from piblin file reader class. Also enables custom columns not recommended for MALDI
            results, but can be used to import different values.
        """

        column_update_dict = read_kwargs.get(p.CUSTOM_COLUMNS, None)

        all_sheets = pd.read_excel(filepath, sheet_name=None)

        file_contents = {}
        for sheet_name, df in all_sheets.items():
            if len(df) == 0:
                continue
            if column_update_dict is None:
                column_update_dict = p.COLUMN_UPDATE_DICT

            skip = 0
            while not all([x in df.columns for x in COLUMN_HEADERS]):
                skip += 1
                df = pd.read_excel(filepath, sheet_name=sheet_name, header=skip)

            sheet_contents = {}
            for column in df.columns:
                if column.lower() in column_update_dict.keys():
                    data_key = column_update_dict[column.lower()]
                    data_value = df[column].to_numpy()

                    sheet_contents[data_key] = data_value
            file_contents[sheet_name] = sheet_contents

        return file_contents

    @property
    def default_mode(self) -> str:
        return ''

    @classmethod
    def _data_from_file_contents(cls, file_contents: Dict, file_location=None, file_name=None,
                                 **read_kwargs) -> MeasurementSet:
        measurements = []
        for sheet_name, contents in file_contents.items():
            mass = contents.get(p.MASS_LABEL, None)
            mz = contents.get(p.MZ_LABEL, None)
            rt = contents.get(p.RT_LABEL, None)
            abundance = contents.get(p.ABUNDANCE_LABEL, None)

            result = FractionalMRDataset(
                mass=mass,
                mz=mz,
                rt=rt,
                abundance=abundance
            )

            file_path = os.path.join(file_location, file_name)

            # xlsx sheet name is in place of the file, maintains uniqueness from single file with multiple sheets
            measurements.append(
                Measurement(datasets=[result], conditions={'file_name': sheet_name}, details={'file_path': file_path})
            )

        return MeasurementSet(measurements=measurements, merge_redundant=False)



"""Piblin_kmd: Tools for Kendrick Mass Defect Analysis

This package provides classes and utilities for analyzing mass spectrometry data
using Fractional Mass Residual (FMR) a form of end-group analysis using KMD-like algorithms.
"""

from .fmr_filereaders.fmr_mass_spreadsheet_reader import MassSpreadsheetReader
from .fmr_filereaders.maldi_mass_list_reader import MaldiMassListReader
from .fmr_classes import FractionalMRDataset, FractionalMRMeasurement
from .fmr_parameters import *
from .fmr_transforms import fmr_transforms as transforms



"""FMR data transformation module.

This module provides transformers for processing Fractional Mass Remainder (FMR) data.
Utility and Distance functions:
    convert_formula_to_mass:
        Convert chemical formulas to masses using molmass package.

    ppm_metric:
        Calculate PPM-based metrics for mass comparisons. This approach uses the average mass of two values to calculate
        the error based on PPM and m/z tolerances.

Calculation and cluster transformations:
    FractionalMRTransform:
        Transform measurement data based on FMR values and repeat unit calculations.

    ClusterTransform:
        Cluster mass spectrometry data based on FMR values. DBsCAN is used for clustering, calling the custom PPM
        metric.

Piblin Utility Transformations:
    FilterByClusterSize:
        Filter clusters based on cluster size and abundance

    sort_cluster_by:
        Sort clusters by various criteria

    update_clusters:
        Update cluster order based on mass values and minimum size criteria
"""

from sklearn.cluster import DBSCAN
from typing import List, Union, Dict
from molmass import Formula
import numpy as np
from piblin.data import Measurement, MeasurementSet, Dataset
from piblin.transform import MeasurementSetTransform, DatasetTransform
from .. import fmr_parameters as p
from ..fmr_parameters import DEFAULT_REPEAT_UNITS
from ..fmr_classes.fmr_datasets import FractionalMRDataset as FmrDataset


def convert_formula_to_mass(formula: str):
    """
    checks valid formula based on molmass package and returns monoisotopic value
    """
    try:
        return float(Formula(formula).monoisotopic_mass)
    except Exception:
        raise ValueError(f"Formula {formula} could not be converted to mass")


def ppm_metric(
        x1: np.array,
        x2: np.array,
        repeat_unit: float,
        mz_tolerance: float = 0,
        ppm_tolerance: float = 5,
        circular: bool = True,
):
    """
    calculates distance between two np.arrays of vectors (mass, FMR) points based on PPM and m/z tolerances.

    Parameters
    ----------
    x1 : np.array
        First vector (mass, FMR)
    x2 : np.array
        Second vector (mass, FMR)
    repeat_unit : float
        The repeat unit value used for calculating the error.
    mz_tolerance : float
        m/z tolerance value in Da (default is 0 mz units)
    ppm_tolerance : float
        PPM tolerance value (default is 5 ppm)
    circular : bool
        If True, the FMR values are treated as a circular variable (default is True).
    """
    c1 = ppm_tolerance * 10 ** -6
    c2 = mz_tolerance
    avg_mass = (x1[0] + x2[0]) / 2
    error = (c1 * avg_mass + c2) / repeat_unit

    if circular:
        result = min(abs(x1[1] - x2[1]), abs(1 - abs(x1[1] - x2[1]))) / error
    else:
        result = abs(x1[1] - x2[1]) / error
    return result


def sort_cluster_by(
        clusters: np.array,
        abundance: np.array = None,
):
    """
    Sorts clusters by their abundance or size. If abundance is not provided, it defaults to an array of ones.

    Parameters
    ----------
    clusters : np.array
        Array of cluster labels to be sorted.
    abundance : np.array, optional
        Array of abundance values corresponding to each cluster. Default is None.

    Returns
    -------
    np.array New array of sorted cluster labels.
    """

    if abundance is None:
        abundance = np.ones(len(clusters))

    unique_clusters = list(set(clusters))
    abundance_array = np.transpose([unique_clusters, np.zeros(len(unique_clusters)), np.zeros(len(unique_clusters))])
    for cluster in unique_clusters:
        count = abundance[clusters == cluster].sum()
        abundance_array[abundance_array[:, 0] == cluster, 1] = count
    abundance_array = abundance_array[abundance_array[:, 1].argsort()][::-1]

    abundance_array[:, 2] = [x for x in range(len(unique_clusters))]
    new_clusters = -1 * np.ones(len(clusters))
    for cluster in unique_clusters:
        new_cluster = abundance_array[np.where(abundance_array[:, 0] == cluster), 2]
        new_clusters[np.where(clusters == cluster)] = new_cluster
    return new_clusters


def update_clusters(
        clusters: np.array,
        masses: np.array,
        min_size: int,
        ru: float = None
):
    """
    Updates the cluster labels based on the mass values and minimum size criteria. If the cluster size is less than the
    minimum size, it assigns a new cluster label.

    Parameters
    ----------
    clusters : np.array
        Array of cluster labels to be updated.
    masses : np.array
        Array of mass values corresponding to each cluster.
    min_size : int
        Minimum size threshold for clusters.
    ru : float, optional
        Repeat unit value for mass normalization. Default is None.

    Returns
    -------
    np.array Updated array of cluster labels.
    """
    if ru is None:
        km_values = masses
    else:
        km_values = masses * round(ru) / ru

    new_clusters = clusters
    for cluster in set(clusters):
        if not np.isnan(cluster):
            cluster_km = km_values[clusters == cluster]
            size1 = len(set(np.round(cluster_km)))
            size2 = len(set(np.floor(cluster_km)))

            if min(size1, size2) < min_size:
                new_clusters[clusters == cluster] = -1

    result = np.zeros(shape=clusters.shape, dtype=np.int_)
    c = max(clusters)
    for n, value in enumerate(clusters):
        if value == -1:
            c += 1
            result[n] = c
        else:
            result[n] = value
    return result


class FractionalMRTransform(MeasurementSetTransform):
    """A transform class for calculating Fractional Mass Remainder (FMR) values.

    This transform calculates FMR values for mass spectrometry data by dividing masses by repeat unit values.
    FMR values are calculated as (mass / repeat_unit) % 1, resulting in values between 0 and 1.

    The transform supports:
    - Multiple repeat units specified as formulas or masses
    - Fractional repeat units which are applied to all repeat units (e.g., RU/2, RU/3)
    - Option to add Kendrick Mass Defect (KMD) calculations
    - Default repeat units from package parameters

    """
    def __init__(self, data_independent_parameters: List[object] = None, *args, **kwargs):
        req = 1
        if len(data_independent_parameters) != req:
            raise ValueError(
                f"Incorrect number of data-independent parameter passed to transform (needs {req}): "
                f"{len(data_independent_parameters)} given"
            )

        self._repeat_units_values: Dict[str, float] = data_independent_parameters[0]
        super().__init__(data_independent_parameters, *args, **kwargs)

    @staticmethod
    def create(
            repeat_units: Union[str, float, List, Dict[str, Union[str, float]]] = None,
            fractional_values: Union[int, List[int]] = 1,
            default_list: bool = True,
            kmd: bool = False,
    ):
        """ Method used to create the piblin FractionalMRTransform, which can be applied FMR MeasurementSet.

        This method effectively builds a comprehensive list of repeat units including fractions and optional Kendrick
        Mass Defect calculations.

        Overview:
        1. Creates an empty repeat unit list if no repeat units provided, accepts input as formula (string), repeat
        unit values (float), a list, or dictionary thereof.
        2. Converts repeat units to mass values (using formulas if given), processes default units if enabled.
        3. Validates fractional values (positive integers only) and creates fractional versions of each repeat unit.
        4. Returns new FractionalMRTransform with processed repeat unit dictionary, including optional KMD calculations.

        Parameters
        ----------
        repeat_units : Union[str, float, List, Dict[str, Union[str, float]]], optional
            Chemical formulas or masses to use as repeat units.
        fractional_values : Union[int, List[int]], default=1
            List of fractional values for fractional repeat units.
        default_list : bool, default=True
            Whether to include default repeat units from package parameters.
        kmd : bool, default=False
            Whether to calculate Kendrick Mass Defect values.
        """
        if repeat_units is None:
            repeat_units = []

        # check repeat unit values
        ru_values = {}
        if isinstance(repeat_units, str) or (isinstance(repeat_units, float)):
            repeat_units = [repeat_units]

        if isinstance(repeat_units, list):
            for ru in repeat_units:
                if isinstance(ru, str):
                    ru_str = ru
                    ru: float = convert_formula_to_mass(ru)
                else:
                    ru_str = f"{p.RU_LABEL}{len(ru_values)}"
                ru_values[ru_str] = ru

        elif isinstance(repeat_units, dict):
            # update values to floats only from formulas
            repeat_units = {x: convert_formula_to_mass(y) if isinstance(y, str) else y for x, y in repeat_units.items()}
            ru_values.update(repeat_units)

        if default_list is True:
            default_values: Dict[str, float] = {x: convert_formula_to_mass(x) for x in DEFAULT_REPEAT_UNITS.values()}
            ru_values.update(default_values)

        # check that fractional values are valid
        if isinstance(fractional_values, int):
            fractional_values = [fractional_values]

        fractional_values = list(set([int(x) for x in fractional_values if x > 0]))  # unique values greater than 0

        if len(fractional_values) == 0:
            fractional_values = [1]

        # adds fractional values of each repeat unit to the final list
        final_rus = {}
        temp_dict = ru_values.copy()
        # add fractional values - default 1
        for f_val in fractional_values:
            final_rus.update({f"{k}_{f_val}": v / int(f_val) for k, v in temp_dict.items()})

        # add kmd values if true
        if kmd is True:
            final_rus.update({f"{k}_k": v/round(v) for k, v in temp_dict.items()})

        return FractionalMRTransform(data_independent_parameters=[final_rus])

    def _apply(self, target: MeasurementSet, **kwargs):
        fmr_measurements = []
        for measurement in target.measurements:
            for dataset in measurement.datasets:
                # check if this is an fMR dataset
                if not isinstance(dataset, FmrDataset):
                    raise ValueError("This transform requires a FractionalMRDataset")

                else:
                    for ru_str, ru_value in self._repeat_units_values.items():
                        properties = dict(zip(dataset.data_array_names,dataset.data_arrays))
                        properties.pop(p.MASS_LIST_LABEL)
                        properties.pop(p.CLUSTER_LABEL)
                        properties.pop(p.FMR_LABEL)
                        properties.pop(p.FILTER_LABEL)

                        new_dataset = FmrDataset(
                            **properties
                        )

                        masses = dataset.data_arrays[dataset.data_array_names.index(p.MASS_LIST_LABEL)]
                        fmr_values = np.array([(x / ru_value) % 1 for x in masses])

                        new_dataset.data_arrays[dataset.data_array_names.index(p.FMR_LABEL)] = fmr_values

                        details = measurement.details.copy()
                        details[p.DETAIL_RU_LABEL] = (ru_str, ru_value)

                        conditions = measurement.conditions.copy()
                        conditions[p.CONDITION_RU_LABEL] = ru_value

                        measurement = Measurement(
                            datasets=[new_dataset],
                            details=details,
                            conditions=conditions
                        )

                        fmr_measurements.append(measurement)

        return MeasurementSet(measurements=fmr_measurements, merge_redundant=False)


class ClusterTransform(MeasurementSetTransform):
    """A transform class for clustering mass spectrometry data based on FMR values.

    This transform performs density-based clustering (DBSCAN) on mass/FMR pairs using a custom PPM-based metric.
    Clusters are formed based on mass accuracy (PPM tolerance), m/z tolerance, and minimum cluster size criteria.

    The clustering process:
    1. Sorts data points by FMR values
    2. Applies DBSCAN with custom PPM metric
    3. Updates cluster labels based on size criteria
    4. Sorts clusters by abundance

    Input datasets must be a measurementset composed of FractionalMRDataset instances with FMR values calculated.
    """
    def __init__(self, data_independent_parameters: List[object] = None, *args, **kwargs):
        req = 4
        if len(data_independent_parameters) != req:
            raise ValueError(
                f"Incorrect number of data-independent parameter passed to transform (needs {req}): {len(data_independent_parameters)} given"
            )
        self._align_params = {
            "mz_tolerance": data_independent_parameters[0],
            "ppm_tolerance": data_independent_parameters[1]
        }
        self._min_samples = data_independent_parameters[2]
        self._eps = data_independent_parameters[3]
        super().__init__(data_independent_parameters, *args, **kwargs)

    @staticmethod
    def create(mz_tol: float, ppm_tol: float, min_samples: int, eps: float = 1):
        """ Method used to create the piblin ClusterTransform, which can be applied to FMR MeasurementSet.

        Parameters
        ----------
        mz_tol : float
            m/z tolerance value in Da.
        ppm_tol : float
            PPM tolerance value.
        min_samples : int
            Minimum number of samples required to form a cluster.
        eps : float, default=1
            The maximum relative distance between two samples for one to be considered as in the neighborhood of the
            other.

        """
        return ClusterTransform(data_independent_parameters=[mz_tol, ppm_tol, min_samples, eps])

    def _apply(self, target: MeasurementSet, **kwargs):
        cluster_measurements = []
        for n, measurement in enumerate(target.measurements):
            repeat_unit = measurement.conditions['repeat_unit_value']
            for m, dataset in enumerate(measurement.datasets):
                ru = measurement.conditions[p.CONDITION_RU_LABEL]
                if not isinstance(dataset, FmrDataset):
                    continue
                mass_values = dataset.data_arrays[dataset.data_array_names.index(p.MASS_LIST_LABEL)]
                fmr_values = dataset.data_arrays[dataset.data_array_names.index(p.FMR_LABEL)]
                if p.ABUNDANCE_LABEL in dataset.data_array_names:
                    abundance = dataset.data_arrays[dataset.data_array_names.index(p.ABUNDANCE_LABEL)]
                else:
                    abundance = None

                x = np.array([mass_values, fmr_values]).T

                sorted_index = np.argsort(x[:, 1])
                revert_sorted_index = np.argsort(sorted_index)
                x = x[sorted_index]

                params = self._align_params.copy()
                params['repeat_unit'] = repeat_unit
                dbscan = DBSCAN(eps=self._eps, min_samples=self._min_samples, metric=ppm_metric,
                                metric_params=params)
                dbscan.fit(x)

                labels = dbscan.labels_
                cluster = labels

                # revert sort on cluster and X
                cluster = cluster[revert_sorted_index]
                cluster = update_clusters(clusters=cluster, masses=mass_values, min_size=self._min_samples, ru=ru)
                cluster = sort_cluster_by(cluster, abundance)

                dataset.data_arrays[dataset.data_array_names.index(p.CLUSTER_LABEL)] = cluster

                details = measurement.details
                details['alignment parameters'] = self._align_params

                measurement = Measurement(datasets=[dataset], details=details, conditions=measurement.conditions)
                cluster_measurements.append(measurement)

        return MeasurementSet(measurements=cluster_measurements, merge_redundant=False)


class FilterByClusterSize(DatasetTransform):
    """A transform class for filtering clusters based on size and abundance criteria.

    This transform removes clusters that don't meet minimum size requirements or are in a specified removal list.
    It updates the filter status of points in removed clusters while preserving the original cluster assignments.

    The filtering process:
    1. Checks each cluster's size against minimum sample threshold
    3. Updates filter flags for specified cluster numbers in the FMR datasets

    Input datasets must be FractionalMRDataset instances with cluster assignments.
    """
    def __init__(self, data_independent_parameters: List[object] = None, *args, **kwargs):
        req = 2
        if len(data_independent_parameters) != req:
            raise ValueError(
                f"Incorrect number of data-independent parameter passed to transform (needs {req}): {len(data_independent_parameters)} given"
            )
        self._min_samples = data_independent_parameters[0]
        self._remove_list = data_independent_parameters[1]
        if self._remove_list is None:
            self._remove_list = []
        self._remove_list = np.array(self._remove_list)

        super().__init__(data_independent_parameters, *args, **kwargs)

    @staticmethod
    def create(min_samples: int, remove_list: List[int] = None):
        return FilterByClusterSize(data_independent_parameters=[min_samples, remove_list])

    def _apply(self, target: Dataset, **kwargs):
        if not isinstance(target, FmrDataset):
            return target

        filter_list = target.data_arrays[target.data_array_names.index(p.FILTER_LABEL)]
        clusters = target.data_arrays[target.data_array_names.index(p.CLUSTER_LABEL)]

        unique_clusters = np.unique(clusters)

        remove_list = self._remove_list

        for cluster in unique_clusters:
            if (len(np.where(clusters == cluster)[0]) < self._min_samples) and (cluster not in remove_list):
                remove_list = np.concat([remove_list, np.array([cluster])])
                filter_list[np.where(clusters == cluster)[0]] = 1

        target.data_arrays[target.data_array_names.index(p.FILTER_LABEL)] = filter_list
        return target



"""FMR Datasets and Mesaurements within piblin framework.

This module provides classes for handling Fractional Mass Remainder (FMR) data within the piblin framework.
It includes FractionalMRDataset for storing mass spectrometry data with FMR calculations, and
FractionalMRMeasurement for managing measurements with specific repeat unit information.


__Classes and Methods__

FractionalMRDataset:
    - remove_clusters:
        Remove specific clusters from the dataset
    - to_dict:
        Convert the dataset to a dictionary format
    - _plot_on_axes:
        Base plotting method inherited from OneDimensionalCompositeDataset. Called through piblin visualize methods

FractionalMRMeasurement:
    - plot_mass_rt:
        Visualize mass vs retention time plots for each identified cluster

Methods are called directly or through piblin Measurement or Dataset objects to analyze and visualize FMR data.
"""

import numpy as np
from typing import Union, List, Dict
import matplotlib
import matplotlib.pyplot as plt
from piblin.data import OneDimensionalCompositeDataset, Measurement
from .. import fmr_parameters as p


class FractionalMRDataset(OneDimensionalCompositeDataset):
    """
    A piblin dataset for storing Fractional Mass Remainder (FMR) data.

    Parameters
    ----------
    mass : List[float] | np.array, optional
        List of mass values.
    mz : List[float] | np.array, optional
        List of mass-to-charge ratio values.
    rt : List[float] | np.array, optional
        List of retention time values.
    x : List[float] | np.array, optional
        List of x-coordinates.
    y : List[float] | np.array, optional
        List of y-coordinates.
    abundance : List[float] | np.array, optional
        List of abundance values.
    mass_priority : bool, default True
        Whether to mass_list is based on mass or mz values.
    """
    def __init__(
            self,
            mass: Union[List[float], np.array] = None,
            mz: Union[List[float], np.array] = None,
            rt: Union[List[float], np.array] = None,
            x: Union[List[float], np.array] = None,
            y: Union[List[float], np.array] = None,
            abundance: Union[List[float], np.array] = None,
            mass_priority: bool = True,
    ):
        if mass is None:
            mass = []
        if mz is None:
            mz = []
        if rt is None:
            rt = []
        if x is None:
            x = []
        if y is None:
            y = []
        if abundance is None:
            abundance = []

        # check that mass or mz values exist
        if max(len(mass), len(mz)) == 0:
            raise ValueError("Missing any mass data")

        if (len(mass) > 0) and (mass_priority is True):
            mass_list = mass
        else:
            mass_list = mz

        if len(mass_list) == 0:
            raise ValueError("if mass values are not priority, then provide mz values")

        n_pts = len(mass_list)

        data_arrays = []
        data_array_names = []
        data_array_units = []

        data_list = [mass_list, mass, mz, rt, x, y, abundance, [0] * n_pts, [np.nan] * n_pts, [np.nan] * n_pts]
        data_dict = dict(zip(p.PARAMETER_LABELS, data_list))

        for name, data in data_dict.items():
            # checks if data is equal in length
            if len(data) == n_pts:
                if isinstance(data, list):
                    data = np.array(data)

                data_arrays.append(data)
                data_array_names.append(name)
                data_array_units.append(p.UNITS_DICT[name])

        super().__init__(
            data_arrays=data_arrays,
            data_array_names=data_array_names,
            data_array_units=data_array_units,
            default_independent_name=p.PARAMETER_LABELS[0],
            default_dependent_name=p.PARAMETER_LABELS[-1],
            source=p.SOURCE
        )

    def remove_clusters(
            self,
            clusters: List[int]
    ):
        """
        Remove specific clusters from the dataset by setting their filter values to 1.
        """
        filter_list = self.__getattribute__(p.FILTER_LABEL)
        cluster_list = self.__getattribute__(p.CLUSTER_LABEL)
        for cluster in clusters:
            filter_list[np.where(cluster_list == cluster)[0]] = 1
        self.__setattr__(p.FILTER_LABEL, filter_list)

    def to_dict(self) -> Dict:
        """
        Convert the dataset to a dictionary format with relevant parameters.
        """
        result = {}
        for prop in p.PARAMETER_LABELS:
            value = getattr(self, prop, None)
            if value is not None:
                result[prop] = value
        return result

    def _plot_on_axes(
                self,
                axes: matplotlib.axes.Axes,
                **axes_plotting_kwargs
    ) -> None:
        """
        Updates plotting method for piblin visualize function.
        """
        # get variables
        filter_list = self.data_arrays[self.data_array_names.index(p.FILTER_LABEL)]
        mass_list = self.data_arrays[self.data_array_names.index(p.MASS_LIST_LABEL)]
        fmr_list = self.data_arrays[self.data_array_names.index(p.FMR_LABEL)]
        if p.ABUNDANCE_LABEL in self.data_array_names:
            abundance = self.data_arrays[self.data_array_names.index(p.ABUNDANCE_LABEL)]
        else:
            abundance = np.ones(len(mass_list))  # if no abundance, use ones

        group = self.data_arrays[self.data_array_names.index(p.CLUSTER_LABEL)]

        # Set up color mapping for the clusters
        cmap_value = axes_plotting_kwargs.get('colormap', 'gist_rainbow')
        cmap = plt.get_cmap(cmap_value)
        unique_groups = np.unique(group[np.where(filter_list == 0)[0]])
        # get size of pts
        group_numeric = [np.where(unique_groups == g)[0] for g in group]
        group_numeric = np.array([x[0] if len(x) > 0 else -1 for x in group_numeric])
        n_size = axes_plotting_kwargs.get('cmap_split', 256)
        group_numeric = np.round(group_numeric /n_size) + np.mod(group_numeric, n_size) * n_size
        # get max point size
        size = axes_plotting_kwargs.get('point_size', 7)

        if abundance is not None:
            # normalized to max abundance of unfiltered peaks
            if any([x == 0 for x in filter_list]):
                max_abund = max(abundance[np.where(filter_list == 0)[0]])
            else:
                max_abund = max(abundance)
            normalized_size = 2 * abundance / max_abund * size
        else:
            normalized_size = size

        filter_labels = axes_plotting_kwargs.get('filter_labels', [0, 1])
        # plot points
        for value in filter_labels:
            # Add scatter plot separately
            mass_ = mass_list[np.where(filter_list == value)[0]]
            fmr_ = fmr_list[np.where(filter_list == value)[0]]
            group_ = group_numeric[np.where(filter_list == value)[0]]
            size_ = normalized_size[np.where(filter_list == value)[0]]
            if value == 0:
                axes.scatter(x=mass_, y=fmr_, s=size_, c=group_, cmap=cmap)
            else:
                axes.scatter(x=mass_, y=fmr_, s=size_, facecolors='none', edgecolors='black', alpha=0.5, linewidth=0.25)

        axes.set_ylim(0, 1)


class FractionalMRMeasurement(Measurement):
    """
    A piblin measurement class for Fractional Mass Remainder (FMR) data.
    """
    def __init__(
            self,
            dataset: FractionalMRDataset,
            repeat_unit_information: tuple[str, float],
            details: dict = None,
            conditions: dict = None,
    ):
        if details is None:
            details = {}
            details[p.DETAIL_RU_LABEL] = repeat_unit_information
        if conditions is None:
            conditions = {}
            conditions[p.CONDITION_RU_LABEL] = repeat_unit_information[1]

        super().__init__(
            datasets=[dataset],
            conditions=conditions,
            details=details,
        )

    def plot_mass_rt(self) -> List[plt.Figure]:
        """Plot mass vs retention time for each group in the dataset.

        Returns
        -------
        List[matplotlib.figure.Figure]
            A list of figure objects, one for each identified cluster in the dataset.
        """
        result = []
        dataset = self.datasets[0]
        ru_label = [f'{x:.04f}' if isinstance(x, float) else x for x in dataset.details[p.DETAIL_RU_LABEL]]
        if p.RT_LABEL not in dataset.data_array_names:
            return result
        groups = dataset.data_arrays[dataset.data_array_names.index(p.CLUSTER_LABEL)]
        masses = dataset.data_arrays[dataset.data_array_names.index(p.MASS_LIST_LABEL)]
        rt = dataset.data_arrays[dataset.data_array_names.index(p.RT_LABEL)]

        for group in np.unique(groups):
            fig, ax = plt.subplots(figsize=(8, 6))
            ax.scatter(rt[groups == group], masses[groups == group], s=5)
            ax.set_xlabel('Mass (Da)')
            ax.set_ylabel('RT (min)')
            ax.set_title(f'{ru_label} Group {group}')
            result.append(fig)

        return result



"""A file reader module for importing mass spectrometry data from spreadsheet files.

This module provides functionality to read mass spectrometry data from CSV and Excel files
into FractionalMRDataset objects through piblin FileReader class. It supports custom column mapping for mass, m/z,
retention time, spatial position, and abundance values.

Default supported column headers (case-insensitive):
    - mass: 'mass'
    - m/z: 'mz', 'm/z'
    - abundance: 'abundance', 'intensity', 'area'
    - retention time: 'rt', 'time'
    - x position: 'x_pos'
    - y position: 'y_pos'

Parameters are referenced from fmr_parameters.py
Custom column mappings can be provided via read_kwargs using the CUSTOM_COLUMNS parameter.
"""

import pandas as pd
from typing import Dict
from pathlib import Path

from piblin.dataio.fileio.read.file_reader import FileReader
from piblin.data.data_collections.measurement import Measurement
from piblin.data.data_collections.measurement_set import MeasurementSet

from ..fmr_classes.fmr_datasets import FractionalMRDataset
from .. import fmr_parameters as p


class MassSpreadsheetReader(FileReader):
    """
    A piblin file reader for mass spectrometry data in spreadsheet format (csv, xlsx). read_kwargs enable custom column
    reading.
    """
    supported_extensions = {'csv', 'xlsx'}

    def _read_file_contents(self, filepath: str, **read_kwargs):
        """Read the contents of a file into a single object.

        Parameters
        ----------
        filepath : str
            The path to the file to be read.
        read_kwargs : dict
            used for custom column names using 'mass','mz','rt','x','y','abundance' as keys (check fmr_parameters if
            values change)
        """

        column_update_dict = read_kwargs.get(p.CUSTOM_COLUMNS, None)

        if column_update_dict is None:
            column_update_dict = p.COLUMN_UPDATE_DICT
        else:
            column_update_dict = {x.lower(): y for x, y in column_update_dict.items()}

        file_contents = {}
        extension = Path(filepath).suffix.lower()
        if extension == '.csv':
            df = pd.read_csv(filepath)
        elif extension == '.xlsx':
            df = pd.read_excel(filepath)
        else:
            raise ValueError("file extension incorrect")

        for column in df.columns:
            if column.lower() in [x.lower() for x in column_update_dict.keys()]:
                data_key = column_update_dict[column.lower()]
                data_value = df[column].to_numpy()

                file_contents[data_key] = data_value

        return file_contents

    @property
    def default_mode(self) -> str:
        return ''

    @classmethod
    def _data_from_file_contents(cls, file_contents: Dict, file_location=None, file_name=None,
                                 **read_kwargs) -> MeasurementSet:

        mass = file_contents.get(p.MASS_LABEL, None)
        mz = file_contents.get(p.MZ_LABEL, None)
        rt = file_contents.get(p.RT_LABEL, None)
        abundance = file_contents.get(p.ABUNDANCE_LABEL, None)

        result = FractionalMRDataset(
            mass=mass,
            mz=mz,
            rt=rt,
            abundance=abundance
        )

        # removes extension so the file_name can be used in subsequent steps
        file_name = Path(file_name).with_suffix('')

        measurements = [Measurement(datasets=[result], conditions={'file_name': file_name})]
        return MeasurementSet(measurements=measurements, merge_redundant=False)



"""This module contains constants and default values used in the FMR (Fractional Molecular Repeat) analysis.

This module defines default repeat units, data array labels, units, and other parameters used throughout
the FMR analysis pipeline.

The parameters include:
- Default polymer repeat units
- Column header mappings
- Piblin parameters for data and transforms
"""

# default package repeat units
DEFAULT_REPEAT_UNITS = {
    "PEG": "C2 H4 O",
    "PPG": "C3 H6 O",
    "PTHF": "C4 H8 O",
    "PET": "C10 H8 O4",
    "PE": "C2 H4",
    "PP": "C3 H6",
    "Perfluoro": "C F2",
    "PDMS": "C2 H6 Si O",
    "BPA": "C18 H20 O3",
    "Acrylamide": "C3 H5 N O",
    "Acrylic acid": "C3 H4 O2",
    "Nylon 6 6": "C12 H22 N2 O2",
}

# standardized labels for data arrays
MASS_LABEL = 'mass'
MZ_LABEL = 'mz'
RT_LABEL = 'rt'
X_LABEL = 'x'
Y_LABEL = 'y'
ABUNDANCE_LABEL = 'abundance'
MASS_LIST_LABEL = 'mass_list'
FILTER_LABEL = 'filtered'
FMR_LABEL = 'fmr_list'
CLUSTER_LABEL = 'clusters'

CUSTOM_COLUMNS = 'custom_columns'

# unit names for data arrays
MASS_UNIT = 'Da'
RT_UNIT = 'min'
ABUNDANCE_UNIT = 'AU'
CLUSTER_UNIT = 'count'

# FILEREADER: used to check if columns headers match the following and replaces them with standardized names for data
# arrays
COLUMN_UPDATE_DICT = {
    'mass': MASS_LABEL,
    'mz': MZ_LABEL,
    'm/z': MZ_LABEL,
    'rt': RT_LABEL,
    'retention time': RT_LABEL,
    'abundance': ABUNDANCE_LABEL,
    'intensity': ABUNDANCE_LABEL,
    'Area': ABUNDANCE_LABEL,
    'area': ABUNDANCE_LABEL,
    'x_pos': X_LABEL,
    'y_pos': Y_LABEL
}

ACCEPTED_COLUMN_HEADERS = list(COLUMN_UPDATE_DICT.keys())
RESULT_COLUMN_HEADERS = list(COLUMN_UPDATE_DICT.values())

# DATASET: used to create standardized piblin dataset
PARAMETER_LABELS = [MASS_LIST_LABEL, MASS_LABEL, MZ_LABEL, RT_LABEL, X_LABEL, Y_LABEL, ABUNDANCE_LABEL, FILTER_LABEL,
                    CLUSTER_LABEL, FMR_LABEL]

PARAMETER_UNITS = [MASS_UNIT, MASS_UNIT, MASS_UNIT, RT_UNIT, '', '', ABUNDANCE_UNIT, '', CLUSTER_UNIT, MASS_LABEL]
UNITS_DICT = dict(zip(PARAMETER_LABELS, PARAMETER_UNITS))
SOURCE = 'Spreadsheet containing mass spectral data was read for fractional MR polymer analysis'

# TRANSFORMS: labels for measurements details and conditions
CONDITION_RU_LABEL = 'repeat_unit_value'
DETAIL_RU_LABEL = 'repeat_unit_information'

# default name for repeat units if float is given instead of string
RU_LABEL = "RU"
